/********************************************************************************
*                                                                               *
*                 F i l e  - A s s o c i a t i o n   T a b l e                  *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXFileDict.cpp,v 1.13 2000/10/26 23:25:05 jeroen Exp $                   *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXGIFIcon.h"
#include "FXBMPIcon.h"
#include "FXJPEGIcon.h"
#include "FXPNGIcon.h"
#include "FXXPMIcon.h"
#include "FXFileDict.h"

  
/*
  Notes:

  - FXIconDict should keep path in an FXString; path should be settable
    from outside, rather than automatically from registry; this will make
    FXIconDict a little bit more generally applicable.
  - FXFileDict needs additional fields, e.g. a print command.
  - The associate member function should be virtual so we can overload it.
  - FXFileDict is solely responsible for determining mime-type, and
    duplicate code in FXDirList and FXFileList is eliminated.
  - We will use two different techniques:

      - For directories, we will match "/usr/people/jeroen", then
        try "/people/jeroen", and finally, try "/jeroen" to determine
        directory bindings (Note we pass a "/" in front so we won't
        match a file binding when looking for directories.
        
        This means we can in many cases keep the same icon bindings
        even if some directory tree of a project is moved around in
        the file system.

      - For files, we will try to match the whole name, then try
        the extensions.
        
      - We will try to match "defaultdirbinding" to determine directory type,
        and "defaultfilebinding" for a file, and "defaultexecbinding", for an
        executable to allow bindings to be set for broad categories.

  - We should look into using the mime-database and content-based
    file type detection [I'm not a big fan of this, but it may sometimes
    be necessary].

  - We should at least organize things so that enough info is passed in
    so we can read a fragment of the content and do it (so that means
    the full pathname, and perhaps some flags).

  - The registry format has been extended; it now is:

    command string ';' extension string ';' bigicon [ ':' bigiconopen ] ';' miniicon [ ':' miniiconopen ] ';' mimetype

*/


#define COMMANDLEN   256
#define EXTENSIONLEN 128
#define MIMETYPELEN  64
#define ICONNAMELEN  128

// You can override the default icon locations to search for your
// particular platform by specifying -DDEFAULTICONPATH="path" on
// the command line.
#ifndef DEFAULTICONPATH
#define DEFAULTICONPATH   "~/.foxicons:/usr/local/share/icons:/usr/share/icons"
#endif


/*******************************************************************************/


// Object implementation
FXIMPLEMENT(FXIconDict,FXDict,NULL,0)


// Build icon table
FXIconDict::FXIconDict(FXApp* a):app(a),path(NULL){
  FXTRACE((100,"FXIconDict::FXIconDict\n"));
  }


// Try and load the icon
void *FXIconDict::createData(const void* ptr){
  register const FXchar *ext;  

  // Make sure
  if(!ptr) return NULL;

  // Get icon path
  if(!path){
    path=app->reg().readStringEntry("SETTINGS","iconpath",DEFAULTICONPATH);
    FXTRACE((150,"FXIconDict: iconpath = %s\n",path));
    }

  // Get file extension
  ext=strrchr((const char*)ptr,'.');

  // Determine what type
  if(ext){
    FXIcon *icon=NULL;

    // Create icon of the right type
    if(comparecase(".gif",ext)==0){
      icon=new FXGIFIcon(app);
      }
    else if(comparecase(".bmp",ext)==0){
      icon=new FXBMPIcon(app);
      }
    else if(comparecase(".xpm",ext)==0){
      icon=new FXXPMIcon(app);
      }
// #ifdef HAVE_JPEGLIB_H
//     else if(comparecase(".jpg",ext)==0){
//       icon=new FXJPEGIcon(app);
//       }
// #endif
// #ifdef HAVE_PNG_H
//     else if(comparecase(".png",ext)==0){
//       icon=new FXPNGIcon(app);
//       }
// #endif
    
    
    // Got icon
    if(icon){
      FXchar iconfile[MAXPATHLEN];
      
      // Find icon in the icon directory
      if(fxsearchpath(iconfile,path,(const char*)ptr)){
        FXFileStream str;
        
        FXTRACE((150,"FXIconDict: found icon in = %s\n",iconfile));
  
        // Try open the file
        if(str.open(iconfile,FXStreamLoad)){

          FXTRACE((150,"FXIconDict: loading = %s\n",iconfile));

          // Load it
          icon->loadPixels(str);

          // Done
          str.close();

          return icon;
          }
        }
      
      // Failed, delete the icon
      delete icon;
      }
    }
  return NULL;
  }


// Free the icon
void FXIconDict::deleteData(void* ptr){
  delete ((FXIcon*)ptr);
  }


// Save data
void FXIconDict::save(FXStream& store) const {
  FXDict::save(store);
  FXASSERT(0);
  }


// Load data
void FXIconDict::load(FXStream& store){ 
  FXDict::load(store);
  FXASSERT(0);
  }


// Destructor
FXIconDict::~FXIconDict(){
  FXTRACE((100,"FXIconDict::~FXIconDict\n"));
  clear();
  }


/*******************************************************************************/

// Object implementation
FXIMPLEMENT(FXFileDict,FXDict,NULL,0)


// Construct file extension mapping dictionary
FXFileDict::FXFileDict(FXApp* a):icons(a){
  FXTRACE((100,"FXFileDict::FXFileDict\n"));
  }


// These registry keys are used for default bindings.
const FXchar FXFileDict::defaultExecBinding[]="defaultexecbinding";
const FXchar FXFileDict::defaultDirBinding[]="defaultdirbinding";
const FXchar FXFileDict::defaultFileBinding[]="defaultfilebinding";


// Create new association from extension
void *FXFileDict::createData(const void* ptr){
  FXchar command[COMMANDLEN];
  FXchar extension[EXTENSIONLEN];
  FXchar mimetype[MIMETYPELEN];
  FXchar bigname[ICONNAMELEN];
  FXchar bignameopen[ICONNAMELEN];
  FXchar mininame[ICONNAMELEN];
  FXchar mininameopen[ICONNAMELEN];
  FXFileAssoc *fileassoc;
  const FXchar *p;
  FXchar *q;
  
  FXTRACE((150,"FXFileDict: adding association: %s\n",(FXchar*)ptr));
  
  // Make association record
  fileassoc=new FXFileAssoc;

  // Scan entry
  p=(FXchar*)ptr;
  q=command;
  while(*p && *p!=';' && (p<&command[COMMANDLEN-1])) *q++=*p++;
  *q='\0';
  if(*p==';') p++;
  
  q=extension;
  while(*p && *p!=';' && (p<&command[EXTENSIONLEN-1])) *q++=*p++;
  *q='\0';
  if(*p==';') p++;
  
  q=bigname;
  while(*p && *p!=';' && *p!=':' && (p<&command[ICONNAMELEN-1])) *q++=*p++;
  *q='\0';
  if(*p==':') p++;
  
  q=bignameopen;
  while(*p && *p!=';' && (p<&command[ICONNAMELEN-1])) *q++=*p++;
  *q='\0';
  if(*p==';') p++;
  
  q=mininame;
  while(*p && *p!=';' && *p!=':' && (p<&command[ICONNAMELEN-1])) *q++=*p++;
  *q='\0';
  if(*p==':') p++;
  
  q=mininameopen;
  while(*p && *p!=';' && (p<&command[ICONNAMELEN-1])) *q++=*p++;
  *q='\0';
  if(*p==';') p++;

  q=mimetype;
  while(*p && *p!=';' && (p<&command[MIMETYPELEN-1])) *q++=*p++;
  *q='\0';

  FXTRACE((150,"FXFileDict: command=\"%s\" extension=\"%s\" mimetype=\"%s\" big=\"%s\"|\"%s\" mini=\"%s\"|\"%s\"\n",command,extension,mimetype,bigname,bignameopen,mininame,mininameopen));

  // Initialize association data
  fileassoc->command=command;
  fileassoc->extension=extension;
  fileassoc->bigicon=NULL;
  fileassoc->miniicon=NULL;
  fileassoc->bigiconopen=NULL;
  fileassoc->miniiconopen=NULL;
  fileassoc->mimetype=mimetype;
  fileassoc->dragtype=0;
  fileassoc->flags=0;

  // Insert icons into icon dictionary
  if(bigname[0]){ fileassoc->bigicon=fileassoc->bigiconopen=icons.insert(bigname); }
  if(mininame[0]){ fileassoc->miniicon=fileassoc->miniiconopen=icons.insert(mininame); }

  // Add open icons also; we will fall back on the regular icons in needed
  if(bignameopen[0]){ fileassoc->bigiconopen=icons.insert(bignameopen); }
  if(mininameopen[0]){ fileassoc->miniiconopen=icons.insert(mininameopen); }

  // Return the binding
  return fileassoc;
  }


// Delete association
void FXFileDict::deleteData(void* ptr){
  delete ((FXFileAssoc*)ptr);
  }


// Replace or add file association
FXFileAssoc* FXFileDict::replace(const FXchar* ext,const FXchar* str){
  
  // Replace entry in registry
  getApp()->reg().writeStringEntry("FILETYPES",ext,str);
  
  // Replace record 
  return (FXFileAssoc*)FXDict::replace(ext,str);
  }


// Remove file association
FXFileAssoc* FXFileDict::remove(const FXchar* ext){
  
  // Delete registry entry for this type
  getApp()->reg().deleteEntry("FILETYPES",ext);
  
  // Remove record
  FXDict::remove(ext);
  
  return NULL;
  }


// Find file association
FXFileAssoc* FXFileDict::associate(const FXchar* key){
  register const FXchar *association;
  register FXFileAssoc* record;

  FXTRACE((140,"FXFileDict: trying key: %s\n",key));

  // See if we have an existing record already
  if((record=find(key))!=NULL) return record;
  
  // If not, see if this entry is known in FILETYPES
  association=getApp()->reg().readStringEntry("FILETYPES",key,(FXchar*)NULL);
  
  // If known, make a record for it now
  if(association) return (FXFileAssoc*)FXDict::insert(key,association);
  
  // Not a known file type
  return NULL;
  }



// Find file association from registry
FXFileAssoc* FXFileDict::findFileBinding(const FXchar* pathname){
  register const FXchar* filename=strrchr(pathname,PATHSEP);
  if(!filename) filename=pathname; else filename=filename+1;
  register FXFileAssoc* record;
  FXTRACE((140,"FXFileDict: searching file binding for: %s\n",pathname));
  record=associate(filename);
  if(record) return record;
  filename=strchr(filename,'.');
  while(filename){
    record=associate(filename+1);
    
    // Experimental thumbnail support
//     if(record){
//       FXIcon* icon=NULL;
//       if(record->mimetype=="image/x-xpm"){
//         FXFileStream str;
//         if(str.open(pathname,FXStreamLoad)){
//           icon=new FXXPMIcon(getApp());
//           icon->loadPixels(str);
//           str.close();
//         FXTRACE((140,"FXFileDict: making thumbnail for: %s\n",pathname));
//           if((icon->getWidth()>32) || (icon->getHeight()>32)){
//             if(icon->getWidth()>icon->getHeight()){
//               icon->scale(32,(32*icon->getHeight())/icon->getWidth());
//               }
//             else{
//               icon->scale((32*icon->getWidth())/icon->getHeight(),32);
//               }
//             }
//           record->bigicon=icon;
//           }
//         }
//       }
    if(record) return record;
    filename=strchr(filename+1,'.');
    }
  return associate(defaultFileBinding);
  }


// Find directory association from registry
FXFileAssoc* FXFileDict::findDirBinding(const FXchar* pathname){
  register const FXchar* path=pathname;
  register FXFileAssoc* record;
  FXTRACE((140,"FXFileDict: searching dir binding for: %s\n",pathname));
  while(path){
    record=associate(path);
    if(record) return record;
    path=strchr(path+1,PATHSEP);
    }
  return associate(defaultDirBinding);
  }


// Find executable association from registry
FXFileAssoc* FXFileDict::findExecBinding(const FXchar* pathname){
  FXTRACE((140,"FXFileDict: searching exec binding for: %s\n",pathname));
  return associate(defaultExecBinding);
  }


// Save data
void FXFileDict::save(FXStream& store) const {
  FXDict::save(store);
  FXASSERT(0);
  }


// Load data
void FXFileDict::load(FXStream& store){ 
  FXDict::load(store);
  FXASSERT(0);
  }


// Destructor
FXFileDict::~FXFileDict(){
  FXTRACE((100,"FXFileDict::~FXFileDict\n"));
  clear();
  }
